@@ -27,7 +27,7 @@ class MockedBlockBasedTable : public BlockBasedTable {
   MockedBlockBasedTable(Rep* rep, PartitionedIndexBuilder* pib)
       : BlockBasedTable(rep, /*block_cache_tracer=*/nullptr) {
     // Initialize what Open normally does as much as necessary for the test
-    rep->index_key_includes_seq = pib->seperator_is_key_plus_seq();
+    rep->index_key_includes_seq = pib->separator_is_key_plus_seq();
     rep->index_value_is_full = !pib->get_use_value_delta_encoding();
   }
 };
@@ -58,7 +58,7 @@ class MyPartitionedFilterBlockReader : public PartitionedFilterBlockReader {
 class PartitionedFilterBlockTest
     : public testing::Test,
       virtual public ::testing::WithParamInterface<
-          std::tuple<uint32_t, test::UserDefinedTimestampTestMode>> {
+          std::tuple<uint32_t, test::UserDefinedTimestampTestMode, bool>> {
  public:
   Options options_;
   ImmutableOptions ioptions_;
@@ -70,6 +70,7 @@ class PartitionedFilterBlockTest
   int bits_per_key_;
   size_t ts_sz_;
   bool user_defined_timestamps_persisted_;
+  bool decouple_partitioned_filters;
 
   PartitionedFilterBlockTest() : bits_per_key_(10) {
     auto udt_test_mode = std::get<1>(GetParam());
@@ -85,6 +86,8 @@ class PartitionedFilterBlockTest
         NewBloomFilterPolicy(bits_per_key_, false));
     table_options_.format_version = std::get<0>(GetParam());
     table_options_.index_block_restart_interval = 3;
+    table_options_.decouple_partitioned_filters = decouple_partitioned_filters =
+        std::get<2>(GetParam());
   }
 
   ~PartitionedFilterBlockTest() override = default;
@@ -160,18 +163,24 @@ class PartitionedFilterBlockTest
             FilterBuildingContext(table_options_)),
         table_options_.index_block_restart_interval, !kValueDeltaEncoded,
         p_index_builder, partition_size, ts_sz_,
-        user_defined_timestamps_persisted_);
+        user_defined_timestamps_persisted_, decouple_partitioned_filters);
   }
 
   PartitionedFilterBlockReader* NewReader(
-      PartitionedFilterBlockBuilder* builder, PartitionedIndexBuilder* pib) {
+      PartitionedFilterBlockBuilder* builder, PartitionedIndexBuilder* pib,
+      bool expect_empty = false) {
     BlockHandle bh;
     Status status;
     Slice slice;
     std::unique_ptr<const char[]> filter_data;
     do {
-      slice = builder->Finish(bh, &status, &filter_data);
+      status = builder->Finish(bh, &slice, &filter_data);
       bh = Write(slice);
+      if (expect_empty) {
+        // Ensure most efficient "empty" filter is used
+        EXPECT_OK(status);
+        EXPECT_EQ(0, slice.size());
+      }
     } while (status.IsIncomplete());
 
     constexpr bool skip_filters = false;
@@ -196,26 +205,25 @@ class PartitionedFilterBlockTest
   void VerifyReader(PartitionedFilterBlockBuilder* builder,
                     PartitionedIndexBuilder* pib, bool empty = false) {
     std::unique_ptr<PartitionedFilterBlockReader> reader(
-        NewReader(builder, pib));
+        NewReader(builder, pib, empty));
     // Querying added keys
-    const bool no_io = true;
     std::vector<std::string> keys = PrepareKeys(keys_without_ts, kKeyNum);
     for (const auto& key : keys) {
       auto ikey = InternalKey(key, 0, ValueType::kTypeValue);
       const Slice ikey_slice = Slice(*ikey.rep());
       ASSERT_TRUE(reader->KeyMayMatch(
-          StripTimestampFromUserKey(key, ts_sz_), !no_io, &ikey_slice,
+          StripTimestampFromUserKey(key, ts_sz_), &ikey_slice,
           /*get_context=*/nullptr,
-          /*lookup_context=*/nullptr, ReadOptions()));
+          /*lookup_context=*/nullptr, test::kReadOptionsNoIo));
     }
     {
       // querying a key twice
       auto ikey = InternalKey(keys[0], 0, ValueType::kTypeValue);
       const Slice ikey_slice = Slice(*ikey.rep());
       ASSERT_TRUE(reader->KeyMayMatch(
-          StripTimestampFromUserKey(keys[0], ts_sz_), !no_io, &ikey_slice,
+          StripTimestampFromUserKey(keys[0], ts_sz_), &ikey_slice,
           /*get_context=*/nullptr,
-          /*lookup_context=*/nullptr, ReadOptions()));
+          /*lookup_context=*/nullptr, test::kReadOptionsNoIo));
     }
     // querying missing keys
     std::vector<std::string> missing_keys =
@@ -225,15 +233,15 @@ class PartitionedFilterBlockTest
       const Slice ikey_slice = Slice(*ikey.rep());
       if (empty) {
         ASSERT_TRUE(reader->KeyMayMatch(
-            StripTimestampFromUserKey(key, ts_sz_), !no_io, &ikey_slice,
+            StripTimestampFromUserKey(key, ts_sz_), &ikey_slice,
             /*get_context=*/nullptr,
-            /*lookup_context=*/nullptr, ReadOptions()));
+            /*lookup_context=*/nullptr, test::kReadOptionsNoIo));
       } else {
         // assuming a good hash function
         ASSERT_FALSE(reader->KeyMayMatch(
-            StripTimestampFromUserKey(key, ts_sz_), !no_io, &ikey_slice,
+            StripTimestampFromUserKey(key, ts_sz_), &ikey_slice,
             /*get_context=*/nullptr,
-            /*lookup_context=*/nullptr, ReadOptions()));
+            /*lookup_context=*/nullptr, test::kReadOptionsNoIo));
       }
     }
   }
@@ -306,19 +314,22 @@ class PartitionedFilterBlockTest
     std::string key =
         std::string(*InternalKey(user_key, 0, ValueType::kTypeValue).rep());
     BlockHandle dont_care_block_handle(1, 1);
-    builder->AddIndexEntry(&key, nullptr, dont_care_block_handle);
+    std::string scratch;
+    builder->AddIndexEntry(key, nullptr, dont_care_block_handle, &scratch,
+                           false);
   }
 
   void CutABlock(PartitionedIndexBuilder* builder, const std::string& user_key,
                  const std::string& next_user_key) {
     // Assuming a block is cut, add an entry to the index
-    std::string key =
-        std::string(*InternalKey(user_key, 0, ValueType::kTypeValue).rep());
-    std::string next_key = std::string(
-        *InternalKey(next_user_key, 0, ValueType::kTypeValue).rep());
+    std::string key = *InternalKey(user_key, 0, ValueType::kTypeValue).rep();
+    std::string next_key =
+        *InternalKey(next_user_key, 0, ValueType::kTypeValue).rep();
     BlockHandle dont_care_block_handle(1, 1);
     Slice slice = Slice(next_key.data(), next_key.size());
-    builder->AddIndexEntry(&key, &slice, dont_care_block_handle);
+    std::string scratch;
+    builder->AddIndexEntry(key, &slice, dont_care_block_handle, &scratch,
+                           false);
   }
 
   int CountNumOfIndexPartitions(PartitionedIndexBuilder* builder) {
@@ -337,10 +348,10 @@ class PartitionedFilterBlockTest
 // Format versions potentially intersting to partitioning
 INSTANTIATE_TEST_CASE_P(
     FormatVersions, PartitionedFilterBlockTest,
-    testing::Combine(testing::ValuesIn(std::set<uint32_t>{
-                         2, 3, 4, test::kDefaultFormatVersion,
-                         kLatestFormatVersion}),
-                     testing::ValuesIn(test::GetUDTTestModes())));
+    testing::Combine(
+        testing::ValuesIn(std::set<uint32_t>{
+            2, 3, 4, 5, test::kDefaultFormatVersion, kLatestFormatVersion}),
+        testing::ValuesIn(test::GetUDTTestModes()), testing::Bool()));
 
 TEST_P(PartitionedFilterBlockTest, EmptyBuilder) {
   std::unique_ptr<PartitionedIndexBuilder> pib(NewIndexBuilder());
@@ -389,11 +400,10 @@ TEST_P(PartitionedFilterBlockTest, SamePrefixInMultipleBlocks) {
   for (const auto& key : pkeys) {
     auto ikey = InternalKey(key, 0, ValueType::kTypeValue);
     const Slice ikey_slice = Slice(*ikey.rep());
-    ASSERT_TRUE(reader->PrefixMayMatch(prefix_extractor->Transform(key),
-                                       /*no_io=*/false, &ikey_slice,
-                                       /*get_context=*/nullptr,
-                                       /*lookup_context=*/nullptr,
-                                       ReadOptions()));
+    ASSERT_TRUE(
+        reader->PrefixMayMatch(prefix_extractor->Transform(key), &ikey_slice,
+                               /*get_context=*/nullptr,
+                               /*lookup_context=*/nullptr, ReadOptions()));
   }
   // Non-existent keys but with the same prefix
   const std::string pnonkeys_without_ts[4] = {"p-key9", "p-key11", "p-key21",
@@ -403,11 +413,10 @@ TEST_P(PartitionedFilterBlockTest, SamePrefixInMultipleBlocks) {
   for (const auto& key : pnonkeys) {
     auto ikey = InternalKey(key, 0, ValueType::kTypeValue);
     const Slice ikey_slice = Slice(*ikey.rep());
-    ASSERT_TRUE(reader->PrefixMayMatch(prefix_extractor->Transform(key),
-                                       /*no_io=*/false, &ikey_slice,
-                                       /*get_context=*/nullptr,
-                                       /*lookup_context=*/nullptr,
-                                       ReadOptions()));
+    ASSERT_TRUE(
+        reader->PrefixMayMatch(prefix_extractor->Transform(key), &ikey_slice,
+                               /*get_context=*/nullptr,
+                               /*lookup_context=*/nullptr, ReadOptions()));
   }
 }
 
@@ -444,8 +453,7 @@ TEST_P(PartitionedFilterBlockTest, PrefixInWrongPartitionBug) {
     auto prefix = prefix_extractor->Transform(key);
     auto ikey = InternalKey(key, 0, ValueType::kTypeValue);
     const Slice ikey_slice = Slice(*ikey.rep());
-    ASSERT_TRUE(reader->PrefixMayMatch(prefix,
-                                       /*no_io=*/false, &ikey_slice,
+    ASSERT_TRUE(reader->PrefixMayMatch(prefix, &ikey_slice,
                                        /*get_context=*/nullptr,
                                        /*lookup_context=*/nullptr,
                                        ReadOptions()));
@@ -477,4 +485,4 @@ int main(int argc, char** argv) {
   ROCKSDB_NAMESPACE::port::InstallStackTraceHandler();
   ::testing::InitGoogleTest(&argc, argv);
   return RUN_ALL_TESTS();
-}
\ No newline at end of file
+}
